@@ -43,7 +43,7 @@ WHERE atc.table_name = :object_name
 AND      atc.owner = :owner
 ORDER BY atc.column_id;""",),
 'PROCEDURE': ("""
-              argument_name,             
+              NVL(argument_name, 'Return Value') argument_name,             
 data_type,
 in_out,
 default_value
@@ -51,7 +51,6 @@ FROM all_arguments
 WHERE object_name = :object_name
 AND      owner = :owner
 AND      package_name IS NULL
-AND      argument_name IS NOT NULL
 ORDER BY sequence;""",),    
 'PackageObjects':("""
 SELECT DISTINCT object_name
@@ -111,7 +110,7 @@ AND    index_name = :object_name
 """,)
 }
 descQueries['VIEW'] = descQueries['TABLE']
-descQueries['FUNCTION'] = descQueries['PROCEDURE']
+descQueries['FUNCTION'] = descQueries['PROCEDURE'] 
 
 queries = {
 'resolve': """
@@ -185,7 +184,7 @@ and c1.owner = :owner
 }
 
 import sys, os, re, sqlpython, cx_Oracle, pyparsing
-from cmd2 import Cmd, make_option, options
+from cmd2 import Cmd, make_option, options, Statekeeper
 
 if float(sys.version[:3]) < 2.3:
     def enumerate(lst):
@@ -345,7 +344,8 @@ class sqlpyPlus(sqlpython.sqlpython):
         sqlpython.sqlpython.__init__(self)
         self.binds = CaselessDict()
         self.sqlBuffer = []
-        self.settable = ['maxtselctrows', 'maxfetch', 'autobind', 'failover', 'timeout'] # settables must be lowercase
+        self.settable = ['maxtselctrows', 'maxfetch', 'autobind', 
+                         'failover', 'timeout', 'commit_on_exit'] # settables must be lowercase
         self.stdoutBeforeSpool = sys.stdout
         self.spoolFile = None
         self.autobind = False
@@ -416,6 +416,25 @@ class sqlpyPlus(sqlpython.sqlpython):
                   for row in self.rows]
         return '\n'.join(result)
 
+    html_template = """<html>
+  <head>
+    <title py:content="tblname">Table Name</title>
+  </head>
+  <body>
+    <table py:attr="{'id':tblname}">
+      <tr>
+        <th py:for="colname in colnames">
+          <span py:replace="colname">Column Name</span>
+        </th>
+      </tr>
+      <tr py:for="row in rows">
+        <td py:for="itm in row">
+          <span py:replace="str_or_empty(itm)">Value</span>
+        </td>
+      </tr>
+    </table>
+  </body>
+</html>"""
     def output_as_html_table(self):
         result = ''.join('<th>%s</th>' % c for c in self.colnames)
         result = ['  <tr>\n    %s\n  </tr>' % result]
@@ -429,6 +448,12 @@ class sqlpyPlus(sqlpython.sqlpython):
 </table>''' % (self.tblname, '\n'.join(result))
         return result
 
+    #TODO: use serious templating to make these user-tweakable
+
+    def output_as_markup(self, genshi_template):
+        return None
+        #self.tblname, self.colnames, self.rows
+            
     def output_as_list(self, align):
         result = []
         colnamelen = max(len(colname) for colname in self.colnames) + 1        
@@ -477,18 +502,12 @@ class sqlpyPlus(sqlpython.sqlpython):
             result = sqlpython.pmatrix(self.rows, self.curs.description, self.maxfetch)
         return result
 
-    statementEndPattern = re.compile(r'(.*)(;|\\[gGhtxicCsS])\s*(\d*)\s*$', re.DOTALL | re.MULTILINE)
-    # what about quote-enclosed?
-
-    def findTerminator(self, statement):
-        m = self.statementEndPattern.search(statement)
-        if m:
-            return m.groups()
-        else:
-            return statement, None, None
-
     legalOracle = re.compile('[a-zA-Z_$#]')
 
+    rowlimitPattern = pyparsing.Word(pyparsing.nums)('rowlimit')
+    terminatorPattern = (pyparsing.oneOf('; \\s \\S \\c \\C \\t \\x \\h')    
+                        ^ pyparsing.Literal('\n/\n')) ('terminator') + \
+                        pyparsing.Optional(rowlimitPattern)
     def do_select(self, arg, bindVarsIn=None, override_terminator=None):
         """Fetch rows from a table.
 
@@ -500,19 +519,19 @@ class sqlpyPlus(sqlpython.sqlpython):
         ("help terminators" for details)
         """
         bindVarsIn = bindVarsIn or {}
-        self.query = 'select ' + arg
-        (self.query, terminator, rowlimit) = self.findTerminator(self.query)
+        statement = self.parsed('select ' + arg)
+        self.query = statement.unterminated
         if override_terminator:
-            terminator = override_terminator
-        rowlimit = int(rowlimit or 0)
+            statement['terminator'] = override_terminator
+        statement['rowlimit'] = int(statement.rowlimit or 0)
         try:
             self.varsUsed = findBinds(self.query, self.binds, bindVarsIn)
             self.curs.execute(self.query, self.varsUsed)
-            self.rows = self.curs.fetchmany(min(self.maxfetch, (rowlimit or self.maxfetch)))
+            self.rows = self.curs.fetchmany(min(self.maxfetch, (statement.rowlimit or self.maxfetch)))
             self.desc = self.curs.description
             self.rc = self.curs.rowcount
             if self.rc > 0:
-                self.stdout.write('\n%s\n' % (self.output(terminator, rowlimit)))
+                self.stdout.write('\n%s\n' % (self.output(statement.terminator, statement.rowlimit)))
             if self.rc == 0:
                 print '\nNo rows Selected.\n'
             elif self.rc == 1: 
@@ -535,7 +554,8 @@ class sqlpyPlus(sqlpython.sqlpython):
     def do_pull(self, arg, opts):
         """Displays source code."""
 
-        object_type, owner, object_name = self.resolve(arg.strip(self.terminator).upper())
+        arg = self.parsed(arg).unterminated        
+        object_type, owner, object_name = self.resolve(arg.upper())
         if not object_type:
             return
         self.stdout.write("%s %s.%s\n" % (object_type, owner, object_name))
@@ -549,22 +569,31 @@ class sqlpyPlus(sqlpython.sqlpython):
                 except cx_Oracle.DatabaseError:
                     pass
 
-    @options([make_option('-i', '--insensitive', action='store_true', help='case-insensitive search')])                
+    @options([make_option('-i', '--insensitive', action='store_true', help='case-insensitive search'),
+              make_option('-c', '--col', action='store_true', help='find column'),
+              make_option('-t', '--table', action='store_true', help='find table')])                
     def do_find(self, arg, opts):
-        """Finds argument in source code."""
+        """Finds argument in source code or (with -c) in column definitions."""
 
-        if opts.insensitive:
-            searchfor = "LOWER(text)"
-            arg = arg.lower()
+        arg = self.parsed(arg).unterminated.upper()       
+        if opts.col:
+            self.do_select("owner, table_name, column_name from all_tab_columns where column_name like '%%%s%%'" % (arg))
+        elif opts.table:
+            self.do_select("owner, table_name from all_tables where table_name like '%%%s%%'" % (arg))            
         else:
-            searchfor = "text"
-        self.do_select("* from all_source where %s like '%%%s%%'" % (searchfor, arg))
+            if opts.insensitive:
+                searchfor = "LOWER(text)"
+                arg = arg.lower()
+            else:
+                searchfor = "text"
+            self.do_select("* from all_source where %s like '%%%s%%'" % (searchfor, arg))
 
     @options([make_option('-a','--all',action='store_true',
                           help='Describe all objects (not just my own)')])
     def do_describe(self, arg, opts):
         "emulates SQL*Plus's DESCRIBE"
-        
+
+        arg = self.parsed(arg).unterminated.upper()
         if opts.all:
             which_view = (', owner', 'all')
         else:
@@ -573,7 +602,7 @@ class sqlpyPlus(sqlpython.sqlpython):
         if not arg:
             self.do_select("""object_name, object_type%s FROM %s_objects WHERE object_type IN ('TABLE','VIEW','INDEX') ORDER BY object_name""" % which_view)
             return
-        object_type, owner, object_name = self.resolve(arg.strip(self.terminator).upper())
+        object_type, owner, object_name = self.resolve(arg)
         if not object_type:
             self.do_select("""object_name, object_type%s FROM %s_objects
                            WHERE object_type IN ('TABLE','VIEW','INDEX')
@@ -590,12 +619,13 @@ class sqlpyPlus(sqlpython.sqlpython):
             self.curs.execute(descQueries['PackageObjects'][0], {'package_name':object_name, 'owner':owner})
             packageContents = self.curs.fetchall()
             for (packageObj_name,) in packageContents:
-                print packageObj_name
+                self.stdout.write(packageObj_name + '\n')
                 self.do_select(descQueries['PackageObjArgs'][0],bindVarsIn={'package_name':object_name, 'owner':owner, 'object_name':packageObj_name})
     do_desc = do_describe
 
     def do_deps(self, arg):
-        object_type, owner, object_name = self.resolve(arg.strip(self.terminator).upper())
+        arg = self.parsed(arg).unterminated.upper()        
+        object_type, owner, object_name = self.resolve(arg)
         if object_type == 'PACKAGE BODY':
             q = "and (type != 'PACKAGE BODY' or name != :object_name)'"
             object_type = 'PACKAGE'
@@ -613,7 +643,8 @@ class sqlpyPlus(sqlpython.sqlpython):
 
     def do_comments(self, arg):
         'Prints comments on a table and its columns.'
-        object_type, owner, object_name = self.resolve(arg.strip(self.terminator).upper())
+        arg = self.parsed(arg).unterminated.upper()        
+        object_type, owner, object_name = self.resolve(arg)
         if object_type:
             self.curs.execute(queries['tabComments'],{'table_name':object_name, 'owner':owner})
             self.stdout.write("%s %s.%s: %s\n" % (object_type, owner, object_name, self.curs.fetchone()[0]))
@@ -637,6 +668,7 @@ class sqlpyPlus(sqlpython.sqlpython):
             print 'Could not resolve object %s.' % identifier
             object_type, owner, object_name = '', '', ''
         return object_type, owner, object_name
+        #todo: resolve not finding cwm$ table
 
     def do_resolve(self, arg):
         self.stdout.write(self.resolve(arg)+'\n')
@@ -679,6 +711,7 @@ class sqlpyPlus(sqlpython.sqlpython):
         for n in range(len(args2)):
             query = args2[n]
             fnames.append('compare%s.txt' % n)
+            #TODO: update this terminator-stripping
             if query.rstrip()[-1] != self.terminator: 
                 query = '%s%s' % (query, self.terminator)
             self.onecmd_plus_hooks('%s > %s' % (query, fnames[n]))
@@ -788,57 +821,45 @@ class sqlpyPlus(sqlpython.sqlpython):
             for (var, val) in self.binds.items():
                 print ':%s = %s' % (var, val)
 
+    assignmentScanner = Parser(pyparsing.Literal(':=') ^ '=')
     def do_setbind(self, arg):
-        args = arg.split(None, 2)
-        if len(args) < 2:
+        arg = self.parsed(arg).unterminated
+        try:
+            assigner, startat, endat = self.assignmentScanner.scanner.scanString(arg).next()
+        except StopIteration:
             self.do_print(arg)
-        elif len(args) > 2 and args[1] in ('=',':='):
-            var, val = args[0], args[2]
-            if val[0] == val[-1] == "'" and len(val) > 1:
-                self.binds[var] = val[1:-1]
-                return
+            return
+        var, val = arg[:startat].strip(), arg[endat:].strip()
+        if val[0] == val[-1] == "'" and len(val) > 1:
+            self.binds[var] = val[1:-1]
+            return
+        try:
+            self.binds[var] = int(val)
+            return
+        except ValueError:
             try:
-                self.binds[var] = int(val)
+                self.binds[var] = float(val)
                 return
             except ValueError:
-                try:
-                    self.binds[var] = float(val)
-                    return
-                except ValueError:
-                    try:
-                        self.binds[var] = self.curs.callfunc(val, [])
-                        return
-                    except:
-                        pass
-                    # TODO: when setting 
-
-        print 'Could not parse ', args            
+                statekeeper = Statekeeper(self, ('autobind',))  
+                self.autobind = True
+                self.do_select('%s AS %s FROM dual;' % (val, var))
+                statekeeper.restore()
 
     def do_exec(self, arg):
         if arg[0] == ':':
             self.do_setbind(arg[1:])
         else:
+            arg = self.parsed(arg).unterminated
+            varsUsed = findBinds(arg, self.binds, {})
             try:
-                self.curs.execute('begin\n%s;end;' % arg)
+                self.curs.execute('begin\n%s;end;' % arg, varsUsed)
             except Exception, e:
                 print e
-        '''
-        exec :x := 'box'
-        exec :y := sysdate
-        '''
 
-    '''Works:
-    exec myproc()
-    begin
-      myproc();
-    end;
-    
+    '''
     Fails:
-    select n into :n from test;
-    :d := sysdate
-    :n := myfunc()'''
-    
-    
+    select n into :n from test;'''
     
     def anon_plsql(self, line1):
         lines = [line1]
@@ -861,7 +882,8 @@ class sqlpyPlus(sqlpython.sqlpython):
     #def do_create(self, arg):
     #    self.anon_plsql('create ' + arg)
 
-    @options([make_option('-l', '--long', action='store_true', help='long descriptions')])        
+    @options([make_option('-l', '--long', action='store_true', help='long descriptions'),
+              make_option('-a', '--all', action='store_true', help="all schemas' objects")])        
     def do_ls(self, arg, opts):
         where = ''
         if arg:
@@ -869,17 +891,27 @@ class sqlpyPlus(sqlpython.sqlpython):
                   LIKE '%%%s%%'""" % (arg.upper().replace('*','%'))
         else:
             where = ''
+        if opts.all:
+            owner = 'owner'
+            whose = 'all'
+        else:
+            owner = "'' AS owner"
+            whose = 'user'
         result = []
         statement = '''SELECT object_type, object_name,
-                  status, last_ddl_time
-                  FROM   user_objects %s
-                  ORDER BY object_type, object_name''' % (where)
+                  status, last_ddl_time, %s
+                  FROM   %s_objects %s
+                  ORDER BY object_type, object_name''' % (owner, whose, where)
         self.curs.execute(statement)
-        for (object_type, object_name, status, last_ddl_time) in self.curs.fetchall():
-            if opts.long:
-                result.append('%s\t%s\t%s/%s' % (status, last_ddl_time, object_type, object_name))
+        for (object_type, object_name, status, last_ddl_time, owner) in self.curs.fetchall():
+            if opts.all:
+                qualified_name = '%s.%s' % (owner, object_name)
             else:
-                result.append('%s/%s' % (object_type, object_name))
+                qualified_name = object_name
+            if opts.long:
+                result.append('%s\t%s\t%s/%s' % (status, last_ddl_time, object_type, qualified_name))
+            else:
+                result.append('%s/%s' % (object_type, qualified_name))
         self.stdout.write('\n'.join(result) + '\n')
 
     def do_cat(self, arg):
@@ -919,7 +951,8 @@ class sqlpyPlus(sqlpython.sqlpython):
                 traceback.print_exc(file=sys.stdout)                
 
     def do_refs(self, arg):
-        object_type, owner, object_name = self.resolve(arg.strip(self.terminator).upper())
+        arg = self.parsed(arg).unterminated.upper()        
+        object_type, owner, object_name = self.resolve(arg)
         if object_type == 'TABLE':
             self.do_select(queries['refs'],bindVarsIn={'object_name':object_name, 'owner':owner})
 
